"use strict";
function reversebit(val, bitcnt) {
	let res = 0;
	for (let i = 0; i < bitcnt; i++)
		res |= ((val >> i) & 1) << (bitcnt - i - 1);
	return res;
}
// 2bits represent 1 variant, 1 for true, 2 for false, 3 for dash
// Minterm: UInt32 with 16 variants
function QuineMcCluskey(truthTable, varcnt) {
	function countbits(num) {
		let cnt = 0;
		for (; num > 0; num >>= 1)cnt += num & 1;
		return cnt;
	}
	function iscontain(x, y) {
		return !((~x) & y);
	}
	function isonlydifferone(x, y) {
		let flag = false;
		for (let z = x ^ y; z > 0; z >>= 2)
			if ((z & 0x3) == 0x3) {
				if (flag) return false;
				else flag = true;
			} else if ((z & 0x3) != 0) return false;
		return flag;
	}
	let oneterms = [];
	let groups = new Array(varcnt + 1); for (let i = 0; i <= varcnt; i++)groups[i] = [];
	let groupcheck = new Set();
	for (let i = 0; i < 1 << varcnt; i++) {
		if (truthTable[i] != -1) {
			let minterm = 0;
			for (let j = 0; j < varcnt; j++)
				minterm += i >> j & 1 ? 1 << (j * 2) : 2 << (j * 2);
			if (truthTable[i] == 1)
				oneterms.push(minterm);
			groups[countbits(i)].push(minterm);
			groupcheck.add(minterm);
		}
	}
	for (let should_continue = true; should_continue;) {
		should_continue = false;
		let newgroup = new Array(varcnt + 1); for (let i = 0; i <= varcnt; i++)newgroup[i] = [];
		for (let i = 0; i < varcnt; i++) {
			for (let x of groups[i])
				for (let y of groups[i + 1])
					if (isonlydifferone(x, y)) {
						should_continue = true;
						groupcheck.delete(x);
						groupcheck.delete(y);
						if (!groupcheck.has(x | y)) {
							newgroup[i + 1].push(x | y);
							groupcheck.add(x | y);
						}
					}
		}
		groups = newgroup;
	}
	let terms = Array.from(groupcheck);
	let essential = [];
	groups = new Array(oneterms.length);
	for (let i = 0; i < oneterms.length; i++) {
		groups[i] = [];
		for (let j = 0; j < terms.length; j++)
			if (iscontain(terms[j], oneterms[i]))
				groups[i].push(j);
		if (groups[i].length == 1 && !essential.includes(groups[i][0]))
			essential.push(groups[i][0]);
	}
	let poly = [];// [][]
	Label: for (let i = 0; i < oneterms.length; i++) {
		let s = new Set();
		for (let j of groups[i])
			if (essential.includes(j))
				continue Label;
			else s.add(1n << BigInt(j));
		poly.push(s);
	}
	if (poly.length == 0) return Array.from(essential, v => terms[v]);
	while (poly.length > 1) {
		let p1 = poly.pop(), p2 = poly.pop(), res = new Set();
		for (let t1 of p1)
			for (let t2 of p2)
				res.add(t1 | t2);
		poly.push(res);
	}
	let result = Array.from(essential, v => terms[v]);
	let tmin = 0n, tminlength = 1 << 16;
	for (let t of poly[0]) {
		let tlength = 0;
		for (let i = 0; i < terms.length; i++)
			if (t & (1n << BigInt(i))) tlength++;
		if (tlength < tminlength) tmin = t, tminlength = tlength;
	}
	for (let i = 0; i < terms.length; i++)
		if (tmin & (1n << BigInt(i)))
			result.push(terms[i]);
	return result;
}
const ExpressionMap = [
	["(A)&(!A)", [1, 0, 3]],
	["(!A)&(A|!B)", [1, 2]],
	["(!A)&(A|B)", [1, 2, 3]],
	["(!A|B)&(!A|!B)", [3, 5]],
	["(A)&(!A|!B)", [1, 0, 3, 5]],
	["(A|!B)&(!A|!B)", [0, 3, 5]],
	["(A|B)&(!A|!B)", [0, 1, 3, 5]],
	["(!A|!B)&(!A|!B)", [1, 4, 3, 5]],
	["(A)&(!A|B)", [1, 0]],
	["(!A|B)&(A|!B)", [2]],
	["(A|B)&(!A|B)", [0, 1]],
	["(!A|B)&(!A|B)", []],
	["(A|B)&(A|!B)", [0, 1, 2]],
	["(A|!B)&(A|!B)", [0, 2]],
	["(A|B)&(A|B)", [0, 1, 2, 3]],
	["(B|!B)&(B|!B)", [4, 5]],
];
const inducersForCalculator = ['Ara', 'ATc', 'DAPG', 'Xylose', 'Cumate', '4-HBA'];
class Expression {
	clickable = true;

	originalExpression = "";
	varcnt = 2;
	truthTable = new Int8Array(1 << maxVariableNum);
	cnf = "";
	inducers = [];
	clear() { this.truthTable.fill(0); this.varcnt = 0; this.cnf = ""; this.dnf = ""; this.originalExpression = ""; }
	getCNF() { return this.cnf; }
	getTruthTable() { return this.truthTable; }
	getTruthTable(varindex) { return this.truthTable[varindex]; }
	getOriginalExpression() { return this.originalExpression; }

	InvalidExpression() {
		this.varcnt = 0;
		this.truthTable.fill(0);
		this.cnf = "- INVALID -";
		this.dnf = "";
	}
	setExpression(newexp) {
		this.clear();
		this.originalExpression = newexp;
		let exp = newexp.toUpperCase().trim();
		switch (exp) {
			case "0": case "F": case "FALSE": this.truthTable.fill(-1); this.cnf = this.dnf = '0'; return;
			case "1": case "T": case "TRUE": this.truthTable.fill(1); this.cnf = this.dnf = '1'; return;
			case "": return;
		}

		const validchar = `ABCDEFGH|&!^()`;
		let flag = false;
		for (let c of exp) {
			let ind = validchar.indexOf(c);
			if (ind == -1)
				return this.InvalidExpression();
			if (ind < maxVariableNum) {
				if (flag) return this.InvalidExpression();
				flag = true;
				if (this.varcnt < ind + 1)
					this.varcnt = ind + 1;
			} else flag = false;
		}

		try {
			for (let i = 0; i < maxVariableNum; i++)
				exp = exp.replaceAll(String.fromCharCode(65 + i), `(val>>${i}&1)`);
			let f = new Function('val', "return " + exp);
			for (let i = 0; i < 1 << maxVariableNum; i++) {
				let v = f(reversebit(i, maxVariableNum));
				switch (Number(v)) {
					case 0: this.truthTable[i] = -1; break;
					case 1: this.truthTable[i] = 1; break;
					default: return this.InvalidExpression();
				}
			}
		} catch (error) { return this.InvalidExpression(); }

		this.calcCNFDNF();
	}
	setTruthTable(varindex, value) {
		this.originalExpression = "";
		this.truthTable[varindex] = value;
		this.calcCNFDNF();
	}
	calcCNFDNF() {
		let truth = 0;
		for (let i = 3; i >= 0; i--) {
			truth = (truth << 1) | (this.truthTable[i] == 1 ? 1 : 0);
		}
		this.cnf = ExpressionMap[truth][0];
		this.inducers = ExpressionMap[truth][1];
	}
}

const maxVariableNum = 2;
const TruthValue = []; TruthValue[1] = "1"; TruthValue[-1] = "0"; TruthValue[0] = "x";
let expression = new Expression();

function expressionChanged(newexp) {
	newexp = newexp.toUpperCase();
	expression.setExpression(newexp);

	let ele = document.querySelector("#formalizedExpression");
	ele.innerText = expression.cnf;

	let tbody = document.querySelector("#truthTable>tbody");
	for (let i = 0; i < (1 << 2); i++) {
		let td = tbody.children[i].children[2];
		td.innerText = TruthValue[expression.getTruthTable(i)];
	}
	changeInducerTip();
}
function changeTbValue(td, varindex) {
	varindex = parseInt(varindex);
	if (!expression.clickable) return;
	let newval = 0;
	switch (td.innerText) {
		case "0": newval = 1; break;
		case "1": newval = -1; break;
		default: break;
	}
	td.innerText = TruthValue[newval];
	expression.setTruthTable(varindex, newval);
	let ele = document.querySelector("#formalizedExpression");
	ele.innerText = expression.cnf;
	changeInducerTip();
}
function changeInducerTip() {
	let ele = document.querySelector("#inducerTip");
	let inducers = expression.inducers;
	let text = "No inducer needed";
	if (inducers.length > 0) {
		text = "";
		for (let i of inducers) {
			text += "<li class='i" + inducersForCalculator[i] + "'>" + inducersForCalculator[i] + "</li>";
		}
	}
	ele.innerHTML = text;
}
const container = document.getElementById('main-container');
const ELEMENTS = [
    document.getElementById('row-1'), document.getElementById('row-2'),
    document.getElementById('row-3'), document.getElementById('row-4')
];
const IMAGES = [Array.from(ELEMENTS[1].children), Array.from(ELEMENTS[2].children)];

const getNameByPath = (path) => path.split('/').pop().split('.')[0];
const getPath = (img) => img.className.includes('gap') ?
    'gap' : img.children[0].getAttribute('src');
const getName = (img) => getNameByPath(getPath(img));
const getRevName = (name) => {
    if (name == 'terminator') return 'terminator';
    if (name.includes('-r')) return name.split('-r')[0];
    return name + '-r';
};
const getRevImgName = (img) => getRevName(getName(img));
const getRevImgPath = (img) => 'https://static.igem.wiki/teams/5276/software/assets/' + getRevImgName(img) + '.png';
const revImg = (img) => img.children[0].setAttribute("src", getRevImgPath(img));

const rearrImg = (row, indexes) => {
    IMAGES[row].forEach(img => {
        img.style.transformOrigin = '';
        img.style.transform = '';
    });
    const rowContainer = document.getElementById(`row-${row + 2}`);
    rowContainer.innerHTML = '';
    for (let i = 0; i < indexes.length; i++)
        // if (circuits[row][indexes[i]] == -1)
        rowContainer.appendChild(IMAGES[row][indexes[i]]);
};

const highlight = (img, time) => {
    img.style.transition = 'transform 0.4s, opacity 0.4s';
    img.style.transform = 'scale(1.1)';
    img.style.opacity = 0.5;
    setTimeout(() => {
        img.style.opacity = 1;
        img.style.transform = '';
    }, time / 2);
    setTimeout(() => { img.style.transition = ''; }, time);
};

const idTypes = ['site-', 'promoter-', 'gene-', 'terminator'];
const idType = 16;
const idRow = 8;
const idInput = 4;
const idDir = 1;
const map = [new Map(), new Map()];
let circuits = [];
const initCircuitId = () => {
    for (let i = 0; i < IMAGES.length; i++) {
        const circuit = new Array();
        for (const img of IMAGES[i]) {
            let id = 0;
            if (img.className.includes('gap')) {
                map[i].set('gap', -1);
                circuit.push(-1);
                continue;
            }
            // console.log(img);
            const name = getName(img);
            const type = name.split('-')[0] + '-';
            if (idTypes.indexOf(type) == -1) {
                // console.error(name);
                map[i].set(name, 63);
                circuit.push(-1);
                continue;
            }
            id |= !name.includes('-r');
            if (type == "site-") {
                const num = name.split(type).pop().split('-r')[0];
                if (typeof Number(num) === 'number')
                    id |= (Number(num) - 1) << 1;
            } else {
                id |= !name.includes('-n') << 1;
                id |= !name.includes('a') << 2;
            }
            id |= i << 3;
            id |= idTypes.indexOf(type) << 4;

            // console.log(name, id);
            map[i].set(name, id);
            map[i].set(getRevName(name), id ^ 1);
            circuit.push(id);
        }
        circuits.push(circuit);
    }
    // console.log(circuits);
};
initCircuitId();

const inducers = ['Ara', 'Ara-2', 'ATc', 'Cumate', 'DAPG', 'DAPG-2', 'Xylose', '4-HBA'];
const geneID = (inducer) => inducersForCalculator.indexOf(inducer.split('-2')[0]);
const updateDur = 1600;
let rotateStack = [];
const inducerAdded = document.getElementById('inducerAdded');

let curImages = [
    IMAGES[0].map((img) => IMAGES[0].indexOf(img)),
    IMAGES[1].map((img) => IMAGES[1].indexOf(img))
];

const rotate = (row, start, end) => {
    if (start == -1 || end == -1) return;

    const images = curImages[row].slice(start, end).map(i => IMAGES[row][i]);
    const centerX = images.reduce((sum, img) => sum + img.offsetLeft + img.offsetWidth / 2, 0) / images.length;
    const centerY = images.reduce((sum, img) => sum + img.offsetTop + img.offsetHeight / 2, 0) / images.length;
    // const centerY = container.offsetHeight / 2;

    images.forEach(img => {
        const imgX = img.offsetLeft;
        const imgY = img.offsetTop;
        img.style.transition = 'transform ' + (updateDur / 2) + 'ms';
        img.style.transformOrigin = `${centerX - imgX}px ${centerY - imgY}px`;
        img.style.transform = 'rotate(180deg)';
        rotateStack.push(img);
    });

    curImages[row] = curImages[row].slice(0, start)
        .concat(curImages[row].slice(start, end).reverse())
        .concat(curImages[row].slice(end));

    circuits[row] = circuits[row].slice(0, start)
        .concat(circuits[row].slice(start, end).reverse().map(i => i ^ 1))
        .concat(circuits[row].slice(end));

    setTimeout(() => {
        images.forEach(img => revImg(img));
    }, updateDur / 2);
    setTimeout(() => { })
};

const remove = (row, start, end) => {
    if (start == -1 || end == -1) return;

    const images = curImages[row].slice(start, end).map(i => IMAGES[row][i]);
    const prefix = curImages[row].slice(0, start).map(i => IMAGES[row][i]);
    const moveX = IMAGES[row][curImages[row][end]].offsetLeft -
                  IMAGES[row][curImages[row][start]].offsetLeft;

    images.forEach(img => {
        img.style.transition = 'transform ' + (updateDur / 2) + 'ms' +
            ', opacity ' + (updateDur / 2) + 'ms';
        img.style.transform = 'translateY(60%)';
        img.style.opacity = 0;
    });
    prefix.forEach(img => {
        img.style.transition = 'transform ' + (updateDur / 2) + 'ms';
        img.style.transform = `translateX(${moveX}px)`;
    });

    curImages[row] = curImages[row].slice(0, start)
        .concat(curImages[row].slice(end));

    circuits[row] = circuits[row].slice(0, start)
        .concat(circuits[row].slice(end));
};
const getCurExpr = (row) => {
    let expr = '';
    // check if the LuxI/LuxR on the right is reversed
    const lim = (circuits[row].indexOf(0b100110 | (row << 3)) == -1) ?
        circuits[row].length : circuits[row].indexOf(-1);
    for (let i = 0; i < lim; i++) {
        const id = circuits[row][i];
        if (id != -1 && id & (1 << 4) && ((id >> 3) & 1) == row && id & 1) {
            let atom = ['!A', 'A', '!B', 'B'][(id & 0b111) >> 1];
            expr += (expr == '' ? '' : '|') + atom;
        }
    }
    return expr;
};
const checkUpdate = (inducer) => {
	const id = inducers.indexOf(inducer);
    if (id == -1) return -1;

    const row = id >> 2;
    const isSameID = (ele) => ele >> 1 == id;
    // const circuit = curImages[row].map(
    //     i => map[row].get(getName(IMAGES[row][i]))
    // );
    const start = circuits[row].findIndex(isSameID);
    const end = circuits[row].slice(start + 1).findIndex(isSameID) + start + 1;
    // console.log(start, end);
    if (start == -1 || end == -1 || start == end) {
        enableBtns(runBtn, resetBtn, showBtn);
        return -1;
    }
    if (options[geneID(inducer)].disabled == true) {
        alert("Could not add the same inducer twice!");
        return -1;
    }
	return [id, row, start, end];
}
const updateCircuit = (inducer) => {
	const res = checkUpdate(inducer);
	if (res == -1) return -1;
	const [id, row, start, end] = res;

    if (inducerAdded.innerHTML == 'Nothing.')
        inducerAdded.innerHTML = '';
    if (!inducer.includes('-2')) {
        const newInducer = document.createElement('li');
        newInducer.innerHTML = inducer.split('-2')[0];
        newInducer.setAttribute('class', 'i' + inducer);
        inducerAdded.appendChild(newInducer);
    }
    disableBtns(runBtn, resetBtn, showBtn);
    highlight(ELEMENTS[0].getElementsByClassName('gene')[geneID(inducer)], updateDur);
    curImages[row].slice(start, end + 1).map(i => IMAGES[row][i])
        .forEach(img => highlight(img, updateDur / 2));
    if (circuits[row][start] != circuits[row][end]) {
        setTimeout(() => {
            rotate(row, start, end + 1);
        }, updateDur / 2);
    } else {
        setTimeout(() => {
            remove(row, start, end + 1);
        }, updateDur / 2);
    }

    setTimeout(() => {
        rearrImg(row, curImages[row]);
        if (id == 0 && checkUpdate('Ara-2') != -1) updateCircuit('Ara-2');
        else if (id == 4 && checkUpdate('DAPG-2') != -1) updateCircuit('DAPG-2');
        else {
            enableBtns(runBtn, resetBtn, showBtn);
            options[geneID(inducer)].disabled = true;
        }
    }, updateDur); // 动画时间结束后重新排列`
    // console.log(curImages[row]);

};
const runBtn = document.getElementById('run-btn');
const resetBtn = document.getElementById('reset-btn');
const showBtn = document.getElementById('show-btn');
const disableBtns = (...Btns) => {
    Btns.forEach(btn => btn.disabled = true);
};
const enableBtns = (...Btns) => {
    Btns.forEach(btn => btn.disabled = false);
};
disableBtns(resetBtn); // disable initially

const selector = document.getElementById('inducer-selector');
const options = Array.from(selector.children);
const colors = ['#1770B0', '#A61B29', '#F4BD07', '#7C739F', '#FA8705', '#836B70'];

runBtn.addEventListener('click', function () {
    updateCircuit(selector.value);
});
resetBtn.addEventListener('click', function () {
    // window.location.reload();
    options.forEach(opt => opt.disabled = false);
    inducerAdded.innerHTML = 'Nothing.';
    while (rotateStack.length > 0) {
        let img = rotateStack.pop();
        revImg(img);
    }
    for (let i = 0; i < IMAGES.length; i++) {
        IMAGES[i].forEach(img => {
            img.style.transform = '';
            img.style.transition = '';
            img.style.opacity = 1;
        });
        curImages[i] = IMAGES[i].map(img => IMAGES[i].indexOf(img));
        rearrImg(i, curImages[i]);
    }
    circuits = [];
    initCircuitId();
});
showBtn.addEventListener('click', function () {
    const expr = "(" + getCurExpr(0) + ")&(" + getCurExpr(1) + ")";
    document.querySelector("#expr-input").value = expr;
    expressionChanged(expr);
});

function openPopup() {
    let popupContainer = document.getElementById('popupContainer');
    popupContainer.style.display = 'block';
};

// Function to close the popup
function closePopup() {
    let popupContainer = document.getElementById('popupContainer');
    popupContainer.style.display = 'none';
};
openPopup();